"""Overrides to mypy build manager for direct AST pass through."""

from __future__ import annotations

import ast
import os
from typing import Callable, TYPE_CHECKING, TextIO

from jaclang.compiler.absyntree import AstNode
from jaclang.compiler.passes import Pass
from jaclang.compiler.passes.main.fuse_typeinfo_pass import (
    FuseTypeInfoPass,
)

import mypy.build as myb
import mypy.checkexpr as mycke
import mypy.errors as mye
import mypy.fastparse as myfp
import mypy.nodes as mypy_nodes
from mypy.build import BuildSource
from mypy.build import BuildSourceSet
from mypy.build import FileSystemCache
from mypy.build import Graph
from mypy.build import ModuleNotFound
from mypy.build import PRI_INDIRECT
from mypy.build import Plugin
from mypy.build import SearchPaths
from mypy.build import compute_search_paths
from mypy.build import find_module_simple
from mypy.build import find_module_with_reason
from mypy.build import load_plugins
from mypy.build import process_graph
from mypy.options import Options
from mypy.semanal_main import semantic_analysis_for_scc

if TYPE_CHECKING:
    from mypy.report import Reports  # Avoid unconditional slow import


# All the expression nodes of mypy.
EXPRESSION_NODES = (
    mypy_nodes.AssertTypeExpr,
    mypy_nodes.AssignmentExpr,
    mypy_nodes.AwaitExpr,
    mypy_nodes.BytesExpr,
    mypy_nodes.CallExpr,
    mypy_nodes.CastExpr,
    mypy_nodes.ComparisonExpr,
    mypy_nodes.ComplexExpr,
    mypy_nodes.ConditionalExpr,
    mypy_nodes.DictionaryComprehension,
    mypy_nodes.DictExpr,
    mypy_nodes.EllipsisExpr,
    mypy_nodes.EnumCallExpr,
    mypy_nodes.Expression,
    mypy_nodes.FloatExpr,
    mypy_nodes.GeneratorExpr,
    mypy_nodes.IndexExpr,
    mypy_nodes.IntExpr,
    mypy_nodes.LambdaExpr,
    mypy_nodes.ListComprehension,
    mypy_nodes.ListExpr,
    mypy_nodes.MemberExpr,
    mypy_nodes.NamedTupleExpr,
    mypy_nodes.NameExpr,
    mypy_nodes.NewTypeExpr,
    mypy_nodes.OpExpr,
    mypy_nodes.ParamSpecExpr,
    mypy_nodes.PromoteExpr,
    mypy_nodes.RefExpr,
    mypy_nodes.RevealExpr,
    mypy_nodes.SetComprehension,
    mypy_nodes.SetExpr,
    mypy_nodes.SliceExpr,
    mypy_nodes.StarExpr,
    mypy_nodes.StrExpr,
    mypy_nodes.SuperExpr,
    mypy_nodes.TupleExpr,
    mypy_nodes.TypeAliasExpr,
    mypy_nodes.TypedDictExpr,
    mypy_nodes.TypeVarExpr,
    mypy_nodes.TypeVarTupleExpr,
    mypy_nodes.UnaryExpr,
    mypy_nodes.YieldExpr,
    mypy_nodes.YieldFromExpr,
)


mypy_to_jac_node_map: dict[
    tuple[int, int | None, int | None, int | None], list[AstNode]
] = {}


class BuildManager(myb.BuildManager):
    """Overrides to mypy build manager for direct AST pass through."""

    def __init__(
        self,
        data_dir: str,
        search_paths: SearchPaths,
        ignore_prefix: str,
        source_set: BuildSourceSet,
        reports: Reports | None,
        options: Options,
        version_id: str,
        plugin: Plugin,
        plugins_snapshot: dict[str, str],
        errors: Errors,
        flush_errors: Callable[[str | None, list[str], bool], None],
        fscache: FileSystemCache,
        stdout: TextIO,
        stderr: TextIO,
    ) -> None:
        """Override mypy BuildManager constructor to initialize jac related map."""
        global mypy_to_jac_node_map
        super().__init__(
            data_dir,
            search_paths,
            ignore_prefix,
            source_set,
            reports,
            options,
            version_id,
            plugin,
            plugins_snapshot,
            errors,
            flush_errors,
            fscache,
            stdout,
            stderr,
        )
        mypy_to_jac_node_map = {}

    def parse_file(
        self,
        id: str,
        path: str,
        source: str,
        ignore_errors: bool,
        options: myb.Options,
        ast_override: myb.MypyFile | None = None,
    ) -> myb.MypyFile:
        """Parse the source of a file with the given name.

        Raise CompileError if there is a parse error.
        """
        t0 = myb.time.time()
        if ignore_errors:
            self.errors.ignored_files.add(path)
        tree = (
            ast_override
            if ast_override
            else myb.parse(source, path, id, self.errors, options=options)
        )
        tree._fullname = id
        self.add_stats(
            files_parsed=1,
            modules_parsed=int(not tree.is_stub),
            stubs_parsed=int(tree.is_stub),
            parse_time=myb.time.time() - t0,
        )

        if self.errors.is_blockers():
            self.log("Bailing due to parse errors")
            self.errors.raise_error()

        self.errors.set_file_ignored_lines(path, tree.ignored_lines, ignore_errors)
        return tree


class ExpressionChecker(mycke.ExpressionChecker):
    """Overrides to mypy expression checker for direct AST pass through."""

    def __init__(
        self,
        tc: mycke.mypy.checker.TypeChecker,
        msg: mycke.MessageBuilder,
        plugin: mycke.Plugin,
        per_line_checking_time_ns: dict[int, int],
    ) -> None:
        """Override to mypy expression checker for direct AST pass through."""
        super().__init__(tc, msg, plugin, per_line_checking_time_ns)

    def visit_assert_type_expr(self, e: mycke.AssertTypeExpr) -> mycke.Type:
        """Type check AssertTypeExpr expression."""
        out = super().visit_assert_type_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_assignment_expr(self, e: mycke.AssignmentExpr) -> mycke.Type:
        """Type check AssignmentExpr expression."""
        out = super().visit_assignment_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_await_expr(
        self, e: mycke.AwaitExpr, allow_none_return: bool = False
    ) -> mycke.Type:
        """Type check AwaitExpr expression."""
        out = super().visit_await_expr(e, allow_none_return)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_bytes_expr(self, e: mycke.BytesExpr) -> mycke.Type:
        """Type check BytesExpr expression."""
        out = super().visit_bytes_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_call_expr(
        self, e: mycke.CallExpr, allow_none_return: bool = False
    ) -> mycke.Type:
        """Type check CallExpr expression."""
        out = super().visit_call_expr(e, allow_none_return)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_cast_expr(self, e: mycke.CastExpr) -> mycke.Type:
        """Type check CastExpr expression."""
        out = super().visit_cast_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_comparison_expr(self, e: mycke.ComparisonExpr) -> mycke.Type:
        """Type check ComparisonExpr expression."""
        out = super().visit_comparison_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_complex_expr(self, e: mycke.ComplexExpr) -> mycke.Type:
        """Type check ComplexExpr expression."""
        out = super().visit_complex_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_conditional_expr(
        self, e: mycke.ConditionalExpr, allow_none_return: bool = False
    ) -> mycke.Type:
        """Type check ConditionalExpr expression."""
        out = super().visit_conditional_expr(e, allow_none_return)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_dictionary_comprehension(
        self, e: mycke.DictionaryComprehension
    ) -> mycke.Type:
        """Type check DictionaryComprehension expression."""
        out = super().visit_dictionary_comprehension(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_dict_expr(self, e: mycke.DictExpr) -> mycke.Type:
        """Type check DictExpr expression."""
        out = super().visit_dict_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_enum_call_expr(self, e: mycke.EnumCallExpr) -> mycke.Type:
        """Type check EnumCallExpr expression."""
        out = super().visit_enum_call_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_float_expr(self, e: mycke.FloatExpr) -> mycke.Type:
        """Type check FloatExpr expression."""
        out = super().visit_float_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_generator_expr(self, e: mycke.GeneratorExpr) -> mycke.Type:
        """Type check GeneratorExpr expression."""
        out = super().visit_generator_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_index_expr(self, e: mycke.IndexExpr) -> mycke.Type:
        """Type check IndexExpr expression."""
        out = super().visit_index_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_int_expr(self, e: mycke.IntExpr) -> mycke.Type:
        """Type check IntExpr expression."""
        out = super().visit_int_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_lambda_expr(self, e: mycke.LambdaExpr) -> mycke.Type:
        """Type check LambdaExpr expression."""
        out = super().visit_lambda_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_list_comprehension(self, e: mycke.ListComprehension) -> mycke.Type:
        """Type check ListComprehension expression."""
        out = super().visit_list_comprehension(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_list_expr(self, e: mycke.ListExpr) -> mycke.Type:
        """Type check ListExpr expression."""
        out = super().visit_list_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_member_expr(
        self, e: mycke.MemberExpr, is_lvalue: bool = False
    ) -> mycke.Type:
        """Type check MemberExpr expression."""
        out = super().visit_member_expr(e, is_lvalue)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_name_expr(self, e: mycke.NameExpr) -> mycke.Type:
        """Type check NameExpr expression."""
        out = super().visit_name_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_op_expr(self, e: mycke.OpExpr) -> mycke.Type:
        """Type check OpExpr expression."""
        out = super().visit_op_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_reveal_expr(self, e: mycke.RevealExpr) -> mycke.Type:
        """Type check RevealExpr expression."""
        out = super().visit_reveal_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_set_comprehension(self, e: mycke.SetComprehension) -> mycke.Type:
        """Type check SetComprehension expression."""
        out = super().visit_set_comprehension(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_set_expr(self, e: mycke.SetExpr) -> mycke.Type:
        """Type check SetExpr expression."""
        out = super().visit_set_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_slice_expr(self, e: mycke.SliceExpr) -> mycke.Type:
        """Type check SliceExpr expression."""
        out = super().visit_slice_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_star_expr(self, e: mycke.StarExpr) -> mycke.Type:
        """Type check StarExpr expression."""
        out = super().visit_star_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_str_expr(self, e: mycke.StrExpr) -> mycke.Type:
        """Type check StrExpr expression."""
        out = super().visit_str_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_super_expr(self, e: mycke.SuperExpr) -> mycke.Type:
        """Type check SuperExpr expression."""
        out = super().visit_super_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_tuple_expr(self, e: mycke.TupleExpr) -> mycke.Type:
        """Type check TupleExpr expression."""
        out = super().visit_tuple_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_type_alias_expr(self, e: mycke.TypeAliasExpr) -> mycke.Type:
        """Type check TypeAliasExpr expression."""
        out = super().visit_type_alias_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_type_var_expr(self, e: mycke.TypeVarExpr) -> mycke.Type:
        """Type check TypeVarExpr expression."""
        out = super().visit_type_var_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_type_var_tuple_expr(self, e: mycke.TypeVarTupleExpr) -> mycke.Type:
        """Type check TypeVarTupleExpr expression."""
        out = super().visit_type_var_tuple_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_unary_expr(self, e: mycke.UnaryExpr) -> mycke.Type:
        """Type check UnaryExpr expression."""
        out = super().visit_unary_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_yield_expr(self, e: mycke.YieldExpr) -> mycke.Type:
        """Type check YieldExpr expression."""
        out = super().visit_yield_expr(e)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out

    def visit_yield_from_expr(
        self, e: mycke.YieldFromExpr, allow_none_return: bool = False
    ) -> mycke.Type:
        """Type check YieldFromExpr expression."""
        out = super().visit_yield_from_expr(e, allow_none_return)
        FuseTypeInfoPass.node_type_hash[e] = out
        return out


class State(myb.State):
    """Overrides to mypy state for direct AST pass through."""

    manager: BuildManager
    tree: myb.MypyFile | None = None

    def __init__(
        self,
        id: str | None,
        path: str | None,
        source: str | None,
        manager: BuildManager,
        caller_state: myb.State | None = None,
        caller_line: int = 0,
        ancestor_for: myb.State | None = None,
        root_source: bool = False,
        # If `temporary` is True, this State is being created to just
        # quickly parse/load the tree, without an intention to further
        # process it. With this flag, any changes to external state as well
        # as error reporting should be avoided.
        temporary: bool = False,
        ast_override: myb.MypyFile | None = None,
    ) -> None:
        """Override to mypy state for AST pass through."""
        if not temporary:
            assert id or path or source is not None, "Neither id, path nor source given"
        self.manager = manager
        State.order_counter += 1
        self.order = State.order_counter
        self.caller_state = caller_state
        self.caller_line = caller_line
        if caller_state:
            self.import_context = caller_state.import_context.copy()
            self.import_context.append((caller_state.xpath, caller_line))
        else:
            self.import_context = []
        self.id = id or "__main__"
        self.options = manager.options.clone_for_module(self.id)
        self.early_errors: list[myb.ErrorInfo] = []
        self._type_checker = None
        if not path and source is None:
            assert id is not None
            try:
                path, follow_imports = myb.find_module_and_diagnose(
                    manager,
                    id,
                    self.options,
                    caller_state,
                    caller_line,
                    ancestor_for,
                    root_source,
                    skip_diagnose=temporary,
                )
            except myb.ModuleNotFound:
                if not temporary:
                    manager.missing_modules.add(id)
                raise
            if follow_imports == "silent":
                self.ignore_all = True
        elif path and myb.is_silent_import_module(manager, path) and not root_source:
            self.ignore_all = True
        self.path = path
        if path:
            self.abspath = myb.os.path.abspath(path)
        self.xpath = path or "<string>"
        if path and source is None and self.manager.cache_enabled:
            self.meta = myb.find_cache_meta(self.id, path, manager)
            # TODO: Get mtime if not cached.
            if self.meta is not None:
                self.interface_hash = self.meta.interface_hash
                self.meta_source_hash = self.meta.hash
        if path and source is None and self.manager.fscache.isdir(path):
            source = ""
        self.source = source
        self.add_ancestors()
        self.per_line_checking_time_ns = myb.collections.defaultdict(int)
        t0 = myb.time.time()
        self.meta = myb.validate_meta(
            self.meta, self.id, self.path, self.ignore_all, manager
        )
        self.manager.add_stats(validate_meta_time=myb.time.time() - t0)
        if self.meta:
            # Make copies, since we may modify these and want to
            # compare them to the originals later.
            self.dependencies = list(self.meta.dependencies)
            self.dependencies_set = set(self.dependencies)
            self.suppressed = list(self.meta.suppressed)
            self.suppressed_set = set(self.suppressed)
            all_deps = self.dependencies + self.suppressed
            assert len(all_deps) == len(self.meta.dep_prios)
            self.priorities = dict(zip(all_deps, self.meta.dep_prios))

            assert len(all_deps) == len(self.meta.dep_lines)
            self.dep_line_map = dict(zip(all_deps, self.meta.dep_lines))
            if temporary:
                self.load_tree(temporary=True)
            if not manager.use_fine_grained_cache() and myb.exist_added_packages(
                self.suppressed, manager, self.options
            ):
                # Special case: if there were a previously missing package imported here
                # and it is not present, then we need to re-calculate dependencies.
                # This is to support patterns like this:
                #     from missing_package import missing_module  # type: ignore
                # At first mypy doesn't know that `missing_module` is a module
                # (it may be a variable, a class, or a function), so it is not added to
                # suppressed dependencies. Therefore, when the package with module is added,
                # we need to re-calculate dependencies.
                # NOTE: see comment below for why we skip this in fine grained mode.
                self.parse_file(
                    ast_override=ast_override
                )  # This is safe because the cache is anyway stale.
                self.compute_dependencies()
        else:
            # When doing a fine-grained cache load, pretend we only
            # know about modules that have cache information and defer
            # handling new modules until the fine-grained update.
            if manager.use_fine_grained_cache():
                manager.log(f"Deferring module to fine-grained update {path} ({id})")
                raise myb.ModuleNotFound

            # Parse the file (and then some) to get the dependencies.
            self.parse_file(temporary=temporary, ast_override=ast_override)
            self.compute_dependencies()

    def type_checker(self) -> myb.TypeChecker:
        """Return the type checker for this state."""
        if not self._type_checker:
            assert (
                self.tree is not None
            ), "Internal error: must be called on parsed file only"
            manager = self.manager
            self._type_checker = myb.TypeChecker(
                manager.errors,
                manager.modules,
                self.options,
                self.tree,
                self.xpath,
                manager.plugin,
                self.per_line_checking_time_ns,
            )
            self._type_checker.expr_checker = ExpressionChecker(
                self._type_checker,
                self._type_checker.msg,
                self._type_checker.plugin,
                self.per_line_checking_time_ns,
            )

        return self._type_checker

    def parse_file(
        self, *, temporary: bool = False, ast_override: myb.MypyFile | None = None
    ) -> None:
        """Parse file and run first pass of semantic analysis.

        Everything done here is local to the file. Don't depend on imported
        modules in any way. Also record module dependencies based on imports.
        """
        if self.tree is not None:
            # The file was already parsed (in __init__()).
            return

        manager = self.manager

        # Can we reuse a previously parsed AST? This avoids redundant work in daemon.
        cached = self.id in manager.ast_cache
        modules = manager.modules
        if not cached:
            manager.log(f"Parsing {self.xpath} ({self.id})")
        else:
            manager.log(f"Using cached AST for {self.xpath} ({self.id})")

        t0 = myb.time_ref()

        with self.wrap_context():
            source = self.source
            self.source = None  # We won't need it again.
            if self.path and source is None:
                try:
                    path = manager.maybe_swap_for_shadow_path(self.path)
                    source = myb.decode_python_encoding(manager.fscache.read(path))
                    self.source_hash = manager.fscache.hash_digest(path)
                except OSError as ioerr:
                    # ioerr.strerror differs for os.stat failures between Windows and
                    # other systems, but os.strerror(ioerr.errno) does not, so we use that.
                    # (We want the error messages to be platform-independent so that the
                    # tests have predictable output.)
                    raise myb.CompileError(
                        [
                            "mypy: can't read file '{}': {}".format(
                                self.path, myb.os.strerror(ioerr.errno)
                            )
                        ],
                        module_with_blocker=self.id,
                    ) from ioerr
                except (UnicodeDecodeError, myb.DecodeError) as decodeerr:
                    if self.path.endswith(".pyd"):
                        err = (
                            f"mypy: stubgen does not support .pyd files: '{self.path}'"
                        )
                    else:
                        err = f"mypy: can't decode file '{self.path}': {str(decodeerr)}"
                    raise myb.CompileError(
                        [err], module_with_blocker=self.id
                    ) from decodeerr
            elif self.path and self.manager.fscache.isdir(self.path):
                source = ""
                self.source_hash = ""
            else:
                assert source is not None
                self.source_hash = myb.compute_hash(source)

            self.parse_inline_configuration(source)
            if not cached:
                self.tree = manager.parse_file(
                    self.id,
                    self.xpath,
                    source,
                    self.ignore_all or self.options.ignore_errors,
                    self.options,
                    ast_override=ast_override,
                )

            else:
                # Reuse a cached AST
                self.tree = manager.ast_cache[self.id][0]
                manager.errors.set_file_ignored_lines(
                    self.xpath,
                    self.tree.ignored_lines,
                    self.ignore_all or self.options.ignore_errors,
                )

        self.time_spent_us += myb.time_spent_us(t0)

        if not cached:
            # Make a copy of any errors produced during parse time so that
            # fine-grained mode can repeat them when the module is
            # reprocessed.
            self.early_errors = list(manager.errors.error_info_map.get(self.xpath, []))
        else:
            self.early_errors = manager.ast_cache[self.id][1]

        if not temporary:
            modules[self.id] = self.tree

        if not cached:
            self.semantic_analysis_pass1()

        if not temporary:
            self.check_blockers()

        manager.ast_cache[self.id] = (self.tree, self.early_errors)


class ASTConverter(myfp.ASTConverter):
    """Overrides to mypy AST converter for direct AST pass through."""

    def visit(self, node: ast.AST | None) -> myfp.Any:  # noqa: ANN401
        """Override to mypy AST converter for direct AST pass through."""
        ret = super().visit(node)
        if node and ret:
            self.link_mypy_to_jac_node(node, ret)
        return ret

    def make_argument(
        self,
        arg: ast.arg,
        default: ast.expr | None,
        kind: myfp.ArgKind,
        no_type_check: bool,
        pos_only: bool = False,
    ) -> myfp.Argument:
        """Override to mypy AST converter for direct AST pass through."""
        ret = super().make_argument(arg, default, kind, no_type_check, pos_only)
        self.link_mypy_to_jac_node(arg, ret)
        return ret

    def link_mypy_to_jac_node(
        self, node: ast.AST, ret: myfp.Any  # noqa: ANN401
    ) -> None:
        """Link mypy AST node to Jac AST node."""
        if hasattr(node, "jac_link"):
            for i in range(len(node.jac_link)):
                node.jac_link[i].gen.mypy_ast.append(ret)
            mypy_to_jac_node_map[
                (ret.line, ret.column, ret.end_line, ret.end_column)
            ] = node.jac_link
        # else:
        #     raise Exception("AST node not linked to Jac node")


class Errors(mye.Errors):
    """Overrides to mypy errors for direct AST pass through."""

    def __init__(self, cur_pass: Pass, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Override to mypy errors for direct AST pass through."""
        self.cur_pass = cur_pass
        super().__init__(*args, **kwargs)

    def report(
        self,
        line: int,
        column: int | None,
        message: str,
        code: mye.ErrorCode | None = None,
        *,
        blocker: bool = False,
        severity: str = "error",
        file: str | None = None,
        only_once: bool = False,
        allow_dups: bool = False,
        origin_span: mye.Iterable[int] | None = None,
        offset: int = 0,
        end_line: int | None = None,
        end_column: int | None = None,
    ) -> None:
        """Override to mypy errors for direct AST pass through."""
        super().report(
            line,
            column,
            message,
            code=code,
            blocker=blocker,
            severity=severity,
            file=file,
            only_once=only_once,
            allow_dups=allow_dups,
            origin_span=origin_span,
            offset=offset,
            end_line=end_line,
            end_column=end_column,
        )
        if (line, column, end_line, end_column) in mypy_to_jac_node_map:
            self.cur_pass.warning(
                msg=message,
                node_override=mypy_to_jac_node_map[
                    (line, column, end_line, end_column)
                ][0],
            )


def load_graph(
    sources: list[BuildSource],
    manager: BuildManager,
    old_graph: Graph | None = None,
    new_modules: list[State] | None = None,
) -> Graph:
    """Given some source files, load the full dependency graph.

    If an old_graph is passed in, it is used as the starting point and
    modified during graph loading.

    If a new_modules is passed in, any modules that are loaded are
    added to the list. This is an argument and not a return value
    so that the caller can access it even if load_graph fails.

    As this may need to parse files, this can raise CompileError in case
    there are syntax errors.
    """
    graph: Graph = old_graph if old_graph is not None else {}

    # The deque is used to implement breadth-first traversal.
    # TODO: Consider whether to go depth-first instead.  This may
    # affect the order in which we process files within import cycles.
    new = new_modules if new_modules is not None else []
    entry_points: set[str] = set()
    # Seed the graph with the initial root sources.
    for bs in sources:
        try:
            st = State(
                id=bs.module,
                path=bs.path,
                source=bs.text,
                manager=manager,
                root_source=not bs.followed,
            )
        except ModuleNotFound:
            continue
        if st.id in graph:
            manager.errors.set_file(st.xpath, st.id, manager.options)
            manager.errors.report(
                -1,
                -1,
                f'Duplicate module named "{st.id}" (also at "{graph[st.id].xpath}")',
                blocker=True,
            )
            manager.errors.report(
                -1,
                -1,
                "See https://mypy.readthedocs.io/en/stable/running_mypy.html#mapping-file-paths-to-modules "
                "for more info",
                severity="note",
            )
            manager.errors.report(
                -1,
                -1,
                "Common resolutions include: a) using `--exclude` to avoid checking one of them, "
                "b) adding `__init__.py` somewhere, c) using `--explicit-package-bases` or "
                "adjusting MYPYPATH",
                severity="note",
            )

            manager.errors.raise_error()
        graph[st.id] = st
        new.append(st)
        entry_points.add(bs.module)

    # Note: Running this each time could be slow in the daemon. If it's a problem, we
    # can do more work to maintain this incrementally.
    seen_files = {st.abspath: st for st in graph.values() if st.path}

    # Collect dependencies.  We go breadth-first.
    # More nodes might get added to new as we go, but that's fine.
    for st in new:
        assert st.ancestors is not None
        # Strip out indirect dependencies.  These will be dealt with
        # when they show up as direct dependencies, and there's a
        # scenario where they hurt:
        # - Suppose A imports B and B imports C.
        # - Suppose on the next round:
        #   - C is deleted;
        #   - B is updated to remove the dependency on C;
        #   - A is unchanged.
        # - In this case A's cached *direct* dependencies are still valid
        #   (since direct dependencies reflect the imports found in the source)
        #   but A's cached *indirect* dependency on C is wrong.
        dependencies = [
            dep
            for dep in st.dependencies
            if st.priorities.get(dep) != PRI_INDIRECT and "jaclang.vendor" not in dep
        ]
        if not manager.use_fine_grained_cache():
            # TODO: Ideally we could skip here modules that appeared in st.suppressed
            # because they are not in build with `follow-imports=skip`.
            # This way we could avoid overhead of cloning options in `State.__init__()`
            # below to get the option value. This is quite minor performance loss however.
            added = [dep for dep in st.suppressed if find_module_simple(dep, manager)]
        else:
            # During initial loading we don't care about newly added modules,
            # they will be taken care of during fine grained update. See also
            # comment about this in `State.__init__()`.
            added = []
        for dep in st.ancestors + dependencies + st.suppressed:
            ignored = dep in st.suppressed_set and dep not in entry_points
            if ignored and dep not in added:
                manager.missing_modules.add(dep)
            elif dep not in graph:
                try:
                    if dep in st.ancestors:
                        # TODO: Why not 'if dep not in st.dependencies' ?
                        # Ancestors don't have import context.
                        newst = State(
                            id=dep,
                            path=None,
                            source=None,
                            manager=manager,
                            ancestor_for=st,
                        )
                    else:
                        newst = State(
                            id=dep,
                            path=None,
                            source=None,
                            manager=manager,
                            caller_state=st,
                            caller_line=st.dep_line_map.get(dep, 1),
                        )
                except ModuleNotFound:
                    if dep in st.dependencies_set:
                        st.suppress_dependency(dep)
                else:
                    if newst.path:
                        newst_path = os.path.abspath(newst.path)

                        if newst_path in seen_files:
                            manager.errors.report(
                                -1,
                                0,
                                "Source file found twice under different module names: "
                                '"{}" and "{}"'.format(
                                    seen_files[newst_path].id, newst.id
                                ),
                                blocker=True,
                            )
                            manager.errors.report(
                                -1,
                                0,
                                "See https://mypy.readthedocs.io/en/stable/running"
                                + "_mypy.html#mapping-file-paths-to-modules "
                                "for more info",
                                severity="note",
                            )
                            manager.errors.report(
                                -1,
                                0,
                                "Common resolutions include: a) adding `__init__.py` somewhere, "
                                "b) using `--explicit-package-bases` or adjusting MYPYPATH",
                                severity="note",
                            )
                            manager.errors.raise_error()

                        seen_files[newst_path] = newst

                    assert newst.id not in graph, newst.id
                    graph[newst.id] = newst
                    new.append(newst)  # noqa: B038
            if dep in graph and dep in st.suppressed_set:
                # Previously suppressed file is now visible
                st.add_dependency(dep)
    manager.plugin.set_modules(manager.modules)
    return graph


__all__ = [
    "BuildManager",
    "State",
    "BuildSource",
    "BuildSourceSet",
    "FileSystemCache",
    "find_module_with_reason",
    "compute_search_paths",
    "load_graph",
    "load_plugins",
    "process_graph",
    "Errors",
    "Options",
    "ASTConverter",
    "semantic_analysis_for_scc",
]



"""Jac Language Features."""

from __future__ import annotations

import ast as ast3
import types
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    ParamSpec,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from jaclang.compiler import absyntree as ast
from jaclang.compiler.constant import EdgeDir
from jaclang.compiler.passes.main.pyast_gen_pass import PyastGenPass
from jaclang.runtimelib.constructs import (
    AccessLevel,
    Anchor,
    Architype,
    DSFunc,
    EdgeAnchor,
    EdgeArchitype,
    NodeAnchor,
    NodeArchitype,
    Root,
    WalkerArchitype,
)
from jaclang.runtimelib.context import ExecutionContext

import pluggy

hookspec = pluggy.HookspecMarker("jac")
plugin_manager = pluggy.PluginManager("jac")

T = TypeVar("T")
P = ParamSpec("P")


class JacAccessValidationSpec:
    """Jac Access Validation Specs."""

    @staticmethod
    @hookspec(firstresult=True)
    def elevate_root() -> None:
        """Elevate context root to system_root."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def allow_root(
        architype: Architype, root_id: UUID, level: AccessLevel | int | str
    ) -> None:
        """Allow all access from target root graph to current Architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def disallow_root(
        architype: Architype, root_id: UUID, level: AccessLevel | int | str
    ) -> None:
        """Disallow all access from target root graph to current Architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def unrestrict(architype: Architype, level: AccessLevel | int | str) -> None:
        """Allow everyone to access current Architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def restrict(architype: Architype) -> None:
        """Disallow others to access current Architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def check_read_access(to: Anchor) -> bool:
        """Read Access Validation."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def check_connect_access(to: Anchor) -> bool:
        """Write Access Validation."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def check_write_access(to: Anchor) -> bool:
        """Write Access Validation."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def check_access_level(to: Anchor) -> AccessLevel:
        """Access validation."""
        raise NotImplementedError


class JacNodeSpec:
    """Jac Node Operations."""

    @staticmethod
    @hookspec(firstresult=True)
    def node_dot(node: NodeArchitype, dot_file: Optional[str]) -> str:
        """Generate Dot file for visualizing nodes and edges."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_edges(
        node: NodeAnchor,
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        target_obj: Optional[list[NodeArchitype]],
    ) -> list[EdgeArchitype]:
        """Get edges connected to this node."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def edges_to_nodes(
        node: NodeAnchor,
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        target_obj: Optional[list[NodeArchitype]],
    ) -> list[NodeArchitype]:
        """Get set of nodes connected to this node."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def remove_edge(node: NodeAnchor, edge: EdgeAnchor) -> None:
        """Remove reference without checking sync status."""
        raise NotImplementedError


class JacEdgeSpec:
    """Jac Edge Operations."""

    @staticmethod
    @hookspec(firstresult=True)
    def detach(edge: EdgeAnchor) -> None:
        """Detach edge from nodes."""
        raise NotImplementedError


class JacWalkerSpec:
    """Jac Edge Operations."""

    @staticmethod
    @hookspec(firstresult=True)
    def visit_node(
        walker: WalkerArchitype,
        expr: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:  # noqa: ANN401
        """Jac's visit stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def ignore(
        walker: WalkerArchitype,
        expr: (
            list[NodeArchitype | EdgeArchitype]
            | list[NodeArchitype]
            | list[EdgeArchitype]
            | NodeArchitype
            | EdgeArchitype
        ),
    ) -> bool:
        """Jac's ignore stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def spawn_call(op1: Architype, op2: Architype) -> WalkerArchitype:
        """Invoke data spatial call."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def disengage(walker: WalkerArchitype) -> bool:
        """Jac's disengage stmt feature."""
        raise NotImplementedError


class JacBuiltinSpec:
    """Jac Builtins."""

    @staticmethod
    @hookspec(firstresult=True)
    def dotgen(
        node: NodeArchitype,
        depth: int,
        traverse: bool,
        edge_type: Optional[list[str]],
        bfs: bool,
        edge_limit: int,
        node_limit: int,
        dot_file: Optional[str],
    ) -> str:
        """Print the dot graph."""
        raise NotImplementedError


class JacCmdSpec:
    """Jac CLI command."""

    @staticmethod
    @hookspec
    def create_cmd() -> None:
        """Create Jac CLI cmds."""
        raise NotImplementedError


class JacFeatureSpec(
    JacAccessValidationSpec,
    JacNodeSpec,
    JacEdgeSpec,
    JacWalkerSpec,
    JacBuiltinSpec,
    JacCmdSpec,
):
    """Jac Feature."""

    @staticmethod
    @hookspec(firstresult=True)
    def setup() -> None:
        """Set Class References."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_context() -> ExecutionContext:
        """Get current execution context."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def reset_graph(root: Optional[Root]) -> int:
        """Purge current or target graph."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_object_func() -> Callable[[str], Architype | None]:
        """Get object by id func."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def object_ref(obj: Architype) -> str:
        """Get object's id."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_architype(
        cls: type,
        arch_base: Type[Architype],
        on_entry: list[DSFunc],
        on_exit: list[DSFunc],
    ) -> Type[Architype]:
        """Create a obj architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_obj(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a obj architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_node(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a node architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_root(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a root node architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_edge(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a edge architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_generic_edge(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a generic edge architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def make_walker(
        on_entry: list[DSFunc], on_exit: list[DSFunc]
    ) -> Callable[[type], type]:
        """Create a walker architype."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def impl_patch_filename(
        file_loc: str,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        """Update impl file location."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def jac_import(
        target: str,
        base_path: str,
        absorb: bool,
        cachable: bool,
        mdl_alias: Optional[str],
        override_name: Optional[str],
        lng: Optional[str],
        items: Optional[dict[str, Union[str, Optional[str]]]],
        reload_module: Optional[bool],
    ) -> tuple[types.ModuleType, ...]:
        """Core Import Process."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def create_test(test_fun: Callable) -> Callable:
        """Create a new test."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def run_test(
        filepath: str,
        func_name: Optional[str],
        filter: Optional[str],
        xit: bool,
        maxfail: Optional[int],
        directory: Optional[str],
        verbose: bool,
    ) -> int:
        """Run the test suite in the specified .jac file."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def has_instance_default(gen_func: Callable[[], T]) -> T:
        """Jac's has container default feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def report(expr: Any, custom: bool) -> None:  # noqa: ANN401
        """Jac's report stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def edge_ref(
        node_obj: NodeArchitype | list[NodeArchitype],
        target_obj: Optional[NodeArchitype | list[NodeArchitype]],
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
        edges_only: bool,
    ) -> list[NodeArchitype] | list[EdgeArchitype]:
        """Jac's apply_dir stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def connect(
        left: NodeArchitype | list[NodeArchitype],
        right: NodeArchitype | list[NodeArchitype],
        edge_spec: Callable[[NodeAnchor, NodeAnchor], EdgeArchitype],
        edges_only: bool,
    ) -> list[NodeArchitype] | list[EdgeArchitype]:
        """Jac's connect operator feature.

        Note: connect needs to call assign compr with tuple in op
        """
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def disconnect(
        left: NodeArchitype | list[NodeArchitype],
        right: NodeArchitype | list[NodeArchitype],
        dir: EdgeDir,
        filter_func: Optional[Callable[[list[EdgeArchitype]], list[EdgeArchitype]]],
    ) -> bool:  # noqa: ANN401
        """Jac's disconnect operator feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def assign_compr(
        target: list[T], attr_val: tuple[tuple[str], tuple[Any]]
    ) -> list[T]:
        """Jac's assign comprehension feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_root() -> Root:
        """Jac's root getter."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_root_type() -> Type[Root]:
        """Jac's root getter."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def build_edge(
        is_undirected: bool,
        conn_type: Optional[Type[EdgeArchitype] | EdgeArchitype],
        conn_assign: Optional[tuple[tuple, tuple]],
    ) -> Callable[[NodeAnchor, NodeAnchor], EdgeArchitype]:
        """Jac's root getter."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def save(
        obj: Architype | Anchor,
    ) -> None:
        """Destroy object."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def destroy(
        obj: Architype | Anchor,
    ) -> None:
        """Destroy object."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_semstr_type(
        file_loc: str, scope: str, attr: str, return_semstr: bool
    ) -> Optional[str]:
        """Jac's get_semstr_type stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def obj_scope(file_loc: str, attr: str) -> str:
        """Jac's get_semstr_type feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_sem_type(file_loc: str, attr: str) -> tuple[str | None, str | None]:
        """Jac's get_semstr_type feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def with_llm(
        file_loc: str,
        model: Any,  # noqa: ANN401
        model_params: dict[str, Any],
        scope: str,
        incl_info: list[tuple[str, str]],
        excl_info: list[tuple[str, str]],
        inputs: list[tuple[str, str, str, Any]],
        outputs: tuple,
        action: str,
        _globals: dict,
        _locals: Mapping,
    ) -> Any:  # noqa: ANN401
        """Jac's with_llm stmt feature."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def gen_llm_body(_pass: PyastGenPass, node: ast.Ability) -> list[ast3.AST]:
        """Generate the by LLM body."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def by_llm_call(
        _pass: PyastGenPass,
        model: ast3.AST,
        model_params: dict[str, ast.Expr],
        scope: ast3.AST,
        inputs: Sequence[Optional[ast3.AST]],
        outputs: Sequence[Optional[ast3.AST]] | ast3.Call,
        action: Optional[ast3.AST],
        include_info: list[tuple[str, ast3.AST]],
        exclude_info: list[tuple[str, ast3.AST]],
    ) -> ast3.Call:
        """Return the LLM Call, e.g. _Jac.with_llm()."""
        raise NotImplementedError

    @staticmethod
    @hookspec(firstresult=True)
    def get_by_llm_call_args(_pass: PyastGenPass, node: ast.FuncCall) -> dict:
        """Get the by LLM call args."""
        raise NotImplementedError


plugin_manager.add_hookspecs(JacFeatureSpec)



"""Core constructs for Jac Language."""

from __future__ import annotations

import inspect
from dataclasses import asdict, dataclass, field, fields, is_dataclass
from enum import IntEnum
from functools import cached_property
from logging import getLogger
from pickle import dumps
from types import UnionType
from typing import Any, Callable, ClassVar, Optional, TypeVar
from uuid import UUID, uuid4


logger = getLogger(__name__)

TARCH = TypeVar("TARCH", bound="Architype")
TANCH = TypeVar("TANCH", bound="Anchor")


class AccessLevel(IntEnum):
    """Access level enum."""

    NO_ACCESS = -1
    READ = 0
    CONNECT = 1
    WRITE = 2

    @staticmethod
    def cast(val: int | str | AccessLevel) -> AccessLevel:
        """Cast access level."""
        match val:
            case int():
                return AccessLevel(val)
            case str():
                return AccessLevel[val]
            case _:
                return val


@dataclass
class Access:
    """Access Structure."""

    anchors: dict[str, AccessLevel] = field(default_factory=dict)

    def check(self, anchor: str) -> AccessLevel | None:
        """Validate access."""
        return self.anchors.get(anchor)


@dataclass
class Permission:
    """Anchor Access Handler."""

    all: AccessLevel = AccessLevel.NO_ACCESS
    roots: Access = field(default_factory=Access)


@dataclass
class AnchorReport:
    """Report Handler."""

    id: str
    context: dict[str, Any]


@dataclass(eq=False, repr=False, kw_only=True)
class Anchor:
    """Object Anchor."""

    architype: Architype
    id: UUID = field(default_factory=uuid4)
    root: Optional[UUID] = None
    access: Permission = field(default_factory=Permission)
    persistent: bool = False
    hash: int = 0

    def is_populated(self) -> bool:
        """Check if state."""
        return "architype" in self.__dict__

    def make_stub(self: TANCH) -> TANCH:
        """Return unsynced copy of anchor."""
        if self.is_populated():
            unloaded = object.__new__(self.__class__)
            unloaded.id = self.id
            return unloaded
        return self

    def populate(self) -> None:
        """Retrieve the Architype from db and return."""
        from jaclang.plugin.feature import JacFeature as Jac

        jsrc = Jac.get_context().mem

        if anchor := jsrc.find_by_id(self.id):
            self.__dict__.update(anchor.__dict__)

    def __getattr__(self, name: str) -> object:
        """Trigger load if detects unloaded state."""
        if not self.is_populated():
            self.populate()

            if not self.is_populated():
                raise ValueError(
                    f"{self.__class__.__name__} [{self.id}] is not a valid reference!"
                )

            return getattr(self, name)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has not attribute '{name}'"
        )

    def __getstate__(self) -> dict[str, Any]:  # NOTE: May be better type hinting
        """Serialize Anchor."""
        if self.is_populated():
            unlinked = object.__new__(self.architype.__class__)
            unlinked.__dict__.update(self.architype.__dict__)
            unlinked.__dict__.pop("__jac__", None)

            return {
                "id": self.id,
                "architype": unlinked,
                "root": self.root,
                "access": self.access,
                "persistent": self.persistent,
            }
        else:
            return {"id": self.id}

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Deserialize Anchor."""
        self.__dict__.update(state)

        if self.is_populated() and self.architype:
            self.architype.__jac__ = self
            self.hash = hash(dumps(self))

    def __repr__(self) -> str:
        """Override representation."""
        if self.is_populated():
            attrs = ""
            for f in fields(self):
                if f.name in self.__dict__:
                    attrs += f"{f.name}={self.__dict__[f.name]}, "
            attrs = attrs[:-2]
        else:
            attrs = f"id={self.id}"

        return f"{self.__class__.__name__}({attrs})"

    def report(self) -> AnchorReport:
        """Report Anchor."""
        return AnchorReport(
            id=self.id.hex,
            context=(
                asdict(self.architype)
                if is_dataclass(self.architype) and not isinstance(self.architype, type)
                else {}
            ),
        )

    def __hash__(self) -> int:
        """Override hash for anchor."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Override equal implementation."""
        if isinstance(other, Anchor):
            return self.__class__ is other.__class__ and self.id == other.id

        return False


@dataclass(eq=False, repr=False, kw_only=True)
class NodeAnchor(Anchor):
    """Node Anchor."""

    architype: NodeArchitype
    edges: list[EdgeAnchor]

    def __getstate__(self) -> dict[str, object]:
        """Serialize Node Anchor."""
        state = super().__getstate__()

        if self.is_populated():
            state["edges"] = [edge.make_stub() for edge in self.edges]

        return state


@dataclass(eq=False, repr=False, kw_only=True)
class EdgeAnchor(Anchor):
    """Edge Anchor."""

    architype: EdgeArchitype
    source: NodeAnchor
    target: NodeAnchor
    is_undirected: bool

    def __getstate__(self) -> dict[str, object]:
        """Serialize Node Anchor."""
        state = super().__getstate__()

        if self.is_populated():
            state.update(
                {
                    "source": self.source.make_stub(),
                    "target": self.target.make_stub(),
                    "is_undirected": self.is_undirected,
                }
            )

        return state


@dataclass(eq=False, repr=False, kw_only=True)
class WalkerAnchor(Anchor):
    """Walker Anchor."""

    architype: WalkerArchitype
    path: list[NodeAnchor] = field(default_factory=list)
    next: list[NodeAnchor] = field(default_factory=list)
    ignores: list[NodeAnchor] = field(default_factory=list)
    disengaged: bool = False


@dataclass(eq=False, repr=False, kw_only=True)
class ObjectAnchor(Anchor):
    """Edge Anchor."""

    architype: ObjectArchitype


class Architype:
    """Architype Protocol."""

    _jac_entry_funcs_: ClassVar[list[DSFunc]]
    _jac_exit_funcs_: ClassVar[list[DSFunc]]

    def __init__(self) -> None:
        """Create default architype."""
        self.__jac__ = Anchor(architype=self)

    def __repr__(self) -> str:
        """Override repr for architype."""
        return f"{self.__class__.__name__}"


class NodeArchitype(Architype):
    """Node Architype Protocol."""

    __jac__: NodeAnchor

    def __init__(self) -> None:
        """Create node architype."""
        self.__jac__ = NodeAnchor(architype=self, edges=[])


class EdgeArchitype(Architype):
    """Edge Architype Protocol."""

    __jac__: EdgeAnchor


class WalkerArchitype(Architype):
    """Walker Architype Protocol."""

    __jac__: WalkerAnchor

    def __init__(self) -> None:
        """Create walker architype."""
        self.__jac__ = WalkerAnchor(architype=self)


class ObjectArchitype(Architype):
    """Walker Architype Protocol."""

    __jac__: ObjectAnchor

    def __init__(self) -> None:
        """Create walker architype."""
        self.__jac__ = ObjectAnchor(architype=self)


class GenericEdge(EdgeArchitype):
    """Generic Edge."""


class Root(NodeArchitype):
    """Generic Root Node."""

    def __init__(self) -> None:
        """Create root node."""
        self.__jac__ = NodeAnchor(architype=self, persistent=True, edges=[])


@dataclass(eq=False)
class DSFunc:
    """Data Spatial Function."""

    name: str
    func: Callable[[Any, Any], Any] | None = None

    @cached_property
    def trigger(self) -> type | UnionType | tuple[type | UnionType, ...] | None:
        """Get function parameter annotations."""
        if self.func:
            parameters = inspect.signature(self.func, eval_str=True).parameters
            if len(parameters) >= 2:
                second_param = list(parameters.values())[1]
                ty = second_param.annotation
                return ty if ty != inspect._empty else None
        return None

    def resolve(self, cls: type) -> None:
        """Resolve the function."""
        self.func = getattr(cls, self.name)



"""Test pass module."""

import io
import re
import sys

import jaclang.compiler.absyntree as ast
from jaclang.cli import cli
from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import JacImportPass
from jaclang.compiler.passes.main.fuse_typeinfo_pass import FuseTypeInfoPass
from jaclang.compiler.passes.main.schedules import py_code_gen_typed
from jaclang.utils.test import TestCase


class ImportPassPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pygen_jac_cli(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base.jac"), JacImportPass)
        self.assertFalse(state.errors_had)
        self.assertIn("56", str(state.ir.to_dict()))

    def test_import_auto_impl(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("autoimpl.jac"), JacImportPass)
        num_modules = len(state.ir.get_all_sub_nodes(ast.Module))
        mod_names = [i.name for i in state.ir.get_all_sub_nodes(ast.Module)]
        self.assertEqual(num_modules, 4)
        self.assertIn("getme.impl", mod_names)
        self.assertIn("autoimpl.impl", mod_names)
        self.assertIn("autoimpl.something.else.impl", mod_names)

    def test_import_include_auto_impl(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("incautoimpl.jac"), JacImportPass
        )
        num_modules = len(state.ir.get_all_sub_nodes(ast.Module))
        mod_names = [i.name for i in state.ir.get_all_sub_nodes(ast.Module)]
        self.assertEqual(num_modules, 5)
        self.assertIn("getme.impl", mod_names)
        self.assertIn("autoimpl", mod_names)
        self.assertIn("autoimpl.impl", mod_names)
        self.assertIn("autoimpl.something.else.impl", mod_names)

    def test_annexalbe_by_discovery(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("incautoimpl.jac"), JacImportPass
        )
        count = 0
        for i in state.ir.get_all_sub_nodes(ast.Module):
            if i.name != "autoimpl":
                count += 1
                self.assertEqual(i.annexable_by, self.fixture_abs_path("autoimpl.jac"))
        self.assertEqual(count, 4)

    def test_py_raise_map(self) -> None:
        """Basic test for pass."""
        build = jac_file_to_pass(
            self.fixture_abs_path("py_imp_test.jac"),
            FuseTypeInfoPass,
            schedule=py_code_gen_typed,
        )
        assert isinstance(build.ir, ast.Module)
        p = {
            "math": r"jaclang/vendor/mypy/typeshed/stdlib/math.pyi$",
            "pygame_mock": r"pygame_mock/__init__.pyi$",
            "pygame_mock.color": r"pygame_mock/color.py$",
            "pygame_mock.constants": r"pygame_mock/constants.py$",
            "argparse": r"jaclang/vendor/mypy/typeshed/stdlib/argparse.pyi$",
            "builtins": r"jaclang/vendor/mypy/typeshed/stdlib/builtins.pyi$",
            "pygame_mock.display": r"pygame_mock/display.py$",
            "os": r"jaclang/vendor/mypy/typeshed/stdlib/os/__init__.pyi$",
            "genericpath": r"jaclang/vendor/mypy/typeshed/stdlib/genericpath.pyi$",
        }
        for i in p:
            self.assertIn(i, build.ir.py_info.py_raise_map)
            self.assertRegex(
                re.sub(r".*fixtures/", "", build.ir.py_info.py_raise_map[i]).replace(
                    "\\", "/"
                ),
                p[i],
            )

    def test_py_raised_mods(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("py_imp_test.jac"), schedule=py_code_gen_typed
        )
        for i in list(
            filter(
                lambda x: x.py_info.is_raised_from_py,
                state.ir.get_all_sub_nodes(ast.Module),
            )
        ):
            print(ast.Module.get_href_path(i))
        self.assertEqual(
            len(
                list(
                    filter(
                        lambda x: x.py_info.is_raised_from_py,
                        state.ir.get_all_sub_nodes(ast.Module),
                    )
                )
            ),
            11,  # TODO: Need to only link the modules one time
        )

    def test_double_empty_anx(self) -> None:
        """Test importing python."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.run(self.fixture_abs_path("autoimpl.jac"))
        cli.run(self.fixture_abs_path("autoimpl.jac"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("foo", stdout_value)
        self.assertIn("bar", stdout_value)
        self.assertIn("baz", stdout_value)



"""The Jac Programming Language."""

import inspect
import os
from abc import ABCMeta, abstractmethod as abstract
from dataclasses import dataclass, field as dc_field
from types import ModuleType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Tuple,
    Type,
    TypeGuard,
    TypeVar,
    cast,
    override,
)

from jaclang.plugin.builtin import dotgen, jid, jobj  # noqa: F401
from jaclang.plugin.default import JacFeatureImpl
from jaclang.plugin.feature import JacFeature as Jac, plugin_manager
from jaclang.plugin.spec import EdgeDir

__all__ = [
    # Jac types.
    "Obj",
    "Walker",
    "Node",
    "Edge",
    "JacList",
    "EdgeDir",
    "Root",
    # Decorators.
    "with_entry",
    "with_exit",
    "jac_test",
    "abstract",
    "override",
    # Functions.
    "jac_import",
    "field",
    # Builtin.
    "Jac",
    "root",
    "static",
    "dotgen",
    "jid",
    "jobj",
]


# ----------------------------------------------------------------------------
# Plugin Initialization.
# ----------------------------------------------------------------------------

plugin_manager.register(JacFeatureImpl)
plugin_manager.load_setuptools_entrypoints("jac")

T = TypeVar("T")
S = TypeVar("S")  # S is a subtype of T.

# ----------------------------------------------------------------------------
# Meta classes.
# ----------------------------------------------------------------------------


# Since the meta class of the architypes are changine the base class to it's
# suitable archi type, If a class doesn't have a parent class (ie. by default
# inherit from object) __bases__ will be immutable. So we need to use a dummy
# parent class to make it mutable.
#
# Reference: https://stackoverflow.com/a/9639512/10846399
#
class _ArchiTypeBase:

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialize Jac architype base."""


class JacMeta(ABCMeta):
    """Common metaclass for Jac types."""

    def __new__(  # noqa: D102
        cls,
        name: str,
        bases: Tuple[Type, ...],
        dct: Dict[str, Any],
    ) -> "JacMeta":

        # We have added this "__init__" to the jac base class just to make the type checkers happy.
        # Actually the dataclass decorator will create an __init__ function and assign it here bellow.
        if bases == (_ArchiTypeBase,) and "__init__" in dct:
            del dct["__init__"]

        on_entry, on_exit = [], []
        for func in dct.values():
            if hasattr(func, "__jac_entry"):
                on_entry.append(Jac.DSFunc(func.__name__, func))
            if hasattr(func, "__jac_exit"):
                on_exit.append(Jac.DSFunc(func.__name__, func))

        inst = super().__new__(cls, name, bases, dct)
        inst = dataclass(eq=False)(inst)  # type: ignore [arg-type, assignment]
        inst = inst._MAKE_FN(on_entry, on_exit)(inst)  # type: ignore [assignment, attr-defined]
        return inst


# ----------------------------------------------------------------------------
# Base classes.
# ----------------------------------------------------------------------------


class Obj(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for all the jac object types."""

    _MAKE_FN = Jac.make_obj


class Walker(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for all the jac walker types."""

    _MAKE_FN = Jac.make_walker

    def spawn(self, node: "Node | Root") -> "Walker":
        """Spawn a new node from the walker."""
        return Jac.spawn_call(self, node)  # type: ignore [arg-type, return-value]

    def ignore(
        self,
        expr: """(
        Root
        | Node
        | Edge
        | list[Node | Root | Edge]
        | JacList[Node | Root | Edge]
        )""",
    ) -> bool:
        """Ignore statement."""
        return Jac.ignore(self, expr)  # type: ignore [arg-type]

    def visit(
        self,
        expr: """(
            Root
            | Node
            | Edge
            | list[Node | Root | Edge]
            | JacList[Node | Root | Edge]
        )""",
    ) -> bool:
        """Visit statement."""
        return Jac.visit_node(self, expr)  # type: ignore [arg-type]

    def disengage(self) -> None:
        """Disengage statement."""
        Jac.disengage(self)  # type: ignore [arg-type]


class Node(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for all the jac node types."""

    _MAKE_FN = Jac.make_node

    def spawn(self, archi: Walker) -> "Walker":
        """Spawn a new node from the walker."""
        return Jac.spawn_call(self, archi)  # type: ignore [arg-type, return-value]

    def connect(
        self,
        node: "Node | Root | JacList[Node | Root]",
        edge: "type[Edge] | Edge | None" = None,
        undir: bool = False,
        conn_assign: tuple[tuple, tuple] | None = None,
        edges_only: bool = False,
    ) -> "JacList[Node | Root| Edge]":
        """Connect the current node to another node."""
        # TODO: The above edge type should be reviewed, as the bellow can also take None, Edge, type[Edge].
        ret = Jac.connect(
            left=self,  # type: ignore [arg-type]
            right=node,  # type: ignore [arg-type]
            edge_spec=Jac.build_edge(
                is_undirected=undir, conn_type=edge, conn_assign=conn_assign  # type: ignore [arg-type]
            ),
            edges_only=edges_only,
        )
        return JacList(ret)  # type: ignore [arg-type]

    def disconnect(
        self,
        node: "Node | Root | JacList[Node | Root]",
        edge: "type[Edge] | None" = None,
        dir: EdgeDir = EdgeDir.OUT,
    ) -> bool:
        """Disconnect the current node from the graph."""
        filter_func = None
        if edge:
            filter_func = lambda edges: [  # noqa: E731
                ed for ed in edges if isinstance(ed, edge)
            ]
        return Jac.disconnect(self, node, dir=dir, filter_func=filter_func)  # type: ignore [arg-type]

    def refs(
        self,
        edge: "type[Edge] | None" = None,
        cond: "Callable[[Edge], bool] | None" = None,
        target: "Node | Root | JacList[Node|Root] | None" = None,
        dir: EdgeDir = EdgeDir.OUT,
        edges_only: bool = False,
    ) -> "JacList[Node | Root | Edge]":
        """Return all the connected nodes / edges."""
        filter_func = (
            (
                lambda edges: (  # noqa: E731
                    [ed for ed in edges if isinstance(ed, edge) if not cond or cond(ed)]
                )
            )
            if edge
            else None
        )
        ret = plugin_manager.hook.edge_ref(
            node_obj=self,
            target_obj=target,
            dir=dir,
            filter_func=filter_func,
            edges_only=edges_only,
        )
        return JacList(ret)


class Root(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for jac root type."""

    _MAKE_FN = Jac.make_root

    def spawn(self, archi: Walker) -> "Walker":
        """Spawn a new node from the walker."""
        return Jac.spawn_call(self, archi)  # type: ignore [arg-type, return-value]

    def connect(
        self,
        node: "Node | Root | JacList[Node | Root]",
        edge: "type[Edge] | Edge | None" = None,
        undir: bool = False,
        conn_assign: tuple[tuple, tuple] | None = None,
        edges_only: bool = False,
    ) -> "JacList[Node | Root| Edge]":
        """Connect the current node to another node."""
        # TODO: The above edge type should be reviewed, as the bellow can also take None, Edge, type[Edge].
        ret = Jac.connect(
            left=self,  # type: ignore [arg-type]
            right=node,  # type: ignore [arg-type]
            edge_spec=Jac.build_edge(
                is_undirected=undir, conn_type=edge, conn_assign=conn_assign  # type: ignore [arg-type]
            ),
            edges_only=edges_only,
        )
        return JacList(ret)  # type: ignore [arg-type]

    def disconnect(
        self,
        node: "Node | Root | JacList[Node | Root]",
        edge: "type[Edge] | None" = None,
        dir: EdgeDir = EdgeDir.OUT,
    ) -> bool:
        """Disconnect the current node from the graph."""
        filter_func = None
        if edge:
            filter_func = lambda edges: [  # noqa: E731
                ed for ed in edges if isinstance(ed, edge)
            ]
        return Jac.disconnect(self, node, dir=dir, filter_func=filter_func)  # type: ignore [arg-type]

    def refs(
        self,
        edge: "type[Edge] | None" = None,
        cond: "Callable[[Edge], bool] | None" = None,
        target: "Node | Root | JacList[Node|Root] | None" = None,
        dir: EdgeDir = EdgeDir.OUT,
        edges_only: bool = False,
    ) -> "JacList[Node | Root | Edge]":
        """Return all the connected nodes / edges."""
        filter_func = (
            (
                lambda edges: (  # noqa: E731
                    [ed for ed in edges if isinstance(ed, edge) if not cond or cond(ed)]
                )
            )
            if edge
            else None
        )
        ret = plugin_manager.hook.edge_ref(
            node_obj=self,
            target_obj=target,
            dir=dir,
            filter_func=filter_func,
            edges_only=edges_only,
        )
        return JacList(ret)


class Edge(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for all the jac edge types."""

    _MAKE_FN = Jac.make_edge

    def spawn(self, archi: Walker) -> "Walker":
        """Spawn a new node from the walker."""
        return Jac.spawn_call(self, archi)  # type: ignore [arg-type, return-value]


class GenericEdge(_ArchiTypeBase, metaclass=JacMeta):
    """Base class for jac root type."""

    _MAKE_FN = Jac.make_generic_edge

    def spawn(self, archi: Walker) -> "Walker":
        """Spawn a new node from the walker."""
        return Jac.spawn_call(self, archi)  # type: ignore [arg-type, return-value]


class JacList(Generic[T], list[T]):
    """List with jac methods."""

    # Reuse the methods.
    connect = Node.connect
    disconnect = Node.disconnect
    refs = Node.refs

    def filter(
        self,
        ty: Type[S] | None = None,
        fn: Callable[[T | S], TypeGuard[S]] | None = None,
    ) -> "JacList[S]":
        """Filter comprehension."""
        if ty and fn:
            return JacList([item for item in self if isinstance(item, ty) and fn(item)])
        if ty:
            return JacList([item for item in self if isinstance(item, ty)])
        if fn:
            return JacList(list(filter(fn, self)))
        return cast(JacList[S], self)

    def assign(self, attrs: tuple[str], values: tuple[Any]) -> "JacList[T]":
        """Assign Comprehension."""
        return JacList(Jac.assign_compr(self, (attrs, values)))


# ----------------------------------------------------------------------------
# Decorators.
# ----------------------------------------------------------------------------


def with_entry(func: Callable) -> Callable:
    """Mark a method as jac entry with this decorator."""
    setattr(func, "__jac_entry", True)  # noqa: B010
    return func


def with_exit(func: Callable) -> Callable:
    """Mark a method as jac exit with this decorator."""
    setattr(func, "__jac_exit", True)  # noqa: B010
    return func


# ----------------------------------------------------------------------------
# Functions.
# ----------------------------------------------------------------------------


def jac_import(
    target: str,
    lng: str | None = "jac",
    base_path: str | None = None,
    absorb: bool = False,
    cachable: bool = True,
    alias: str | None = None,
    override_name: str | None = None,
    items: dict[str, str | None] | None = None,
    reload_module: bool | None = False,
) -> tuple[ModuleType, ...]:
    """Import a module."""
    base_path = base_path or os.path.dirname(inspect.stack()[1].filename)
    return Jac.jac_import(
        target=target,
        lng=lng,
        base_path=base_path,
        absorb=absorb,
        cachable=cachable,
        mdl_alias=alias,
        override_name=override_name,
        items=items,
        reload_module=reload_module,
    )


def field(
    value: T = None,  # type: ignore [assignment]
    gen: Callable[[], T] | None = None,
    postinit: bool = False,
) -> T:
    """Set the default value to jac architype dataclass."""
    if postinit:
        return dc_field(init=False)
    gen = gen or (lambda: value)  # noqa: E731
    return Jac.has_instance_default(gen_func=gen)  # type: ignore


# ----------------------------------------------------------------------------
# Builtin.
# ----------------------------------------------------------------------------

jac_test = Jac.create_test
static = ClassVar


def root() -> Root:
    """Get Root."""
    return cast(Root, Jac.get_root())



"""Core constructs for Jac Language."""

from __future__ import annotations

import unittest
from dataclasses import dataclass
from typing import Callable, Optional


class JacTestResult(unittest.TextTestResult):
    """Jac test result class."""

    def __init__(
        self,
        stream,  # noqa
        descriptions,  # noqa
        verbosity: int,
        max_failures: Optional[int] = None,
    ) -> None:
        """Initialize FailFastTestResult object."""
        super().__init__(stream, descriptions, verbosity)  # noqa
        self.failures_count = JacTestCheck.failcount
        self.max_failures = max_failures

    def addFailure(self, test, err) -> None:  # noqa
        """Count failures and stop."""
        super().addFailure(test, err)
        self.failures_count += 1
        if self.max_failures is not None and self.failures_count >= self.max_failures:
            self.stop()

    def stop(self) -> None:
        """Stop the test execution."""
        self.shouldStop = True


class JacTextTestRunner(unittest.TextTestRunner):
    """Jac test runner class."""

    def __init__(self, max_failures: Optional[int] = None, **kwargs) -> None:  # noqa
        """Initialize JacTextTestRunner object."""
        self.max_failures = max_failures
        super().__init__(**kwargs)

    def _makeResult(self) -> JacTestResult:  # noqa
        """Override the method to return an instance of JacTestResult."""
        return JacTestResult(
            self.stream,
            self.descriptions,
            self.verbosity,
            max_failures=self.max_failures,
        )


class JacTestCheck:
    """Jac Testing and Checking."""

    test_case = unittest.TestCase()
    test_suite = unittest.TestSuite()

    @dataclass
    class TestSuite:
        """Test Suite."""

        test_case: unittest.FunctionTestCase
        func_name: str

    test_suite_path: dict[str, list[TestSuite]] = {}

    breaker = False
    failcount = 0

    @staticmethod
    def reset() -> None:
        """Clear the test suite."""
        JacTestCheck.test_case = unittest.TestCase()
        JacTestCheck.test_suite = unittest.TestSuite()
        JacTestCheck.test_suite_path = {}

    @staticmethod
    def run_test(
        xit: bool,
        maxfail: int | None,
        verbose: bool,
        filepath: str | None,
        func_name: str | None,
    ) -> None:
        """Run the test suite."""
        verb = 2 if verbose else 1
        test_suite = JacTestCheck.test_suite

        if filepath and filepath.endswith(".test.jac"):
            filepath = filepath[:-9]
        elif filepath and filepath.endswith(".jac"):
            filepath = filepath[:-4]

        if filepath:
            test_cases = JacTestCheck.test_suite_path.get(filepath)
            if test_cases is not None:
                test_suite = unittest.TestSuite()
                for test_case in test_cases:
                    if func_name:
                        if test_case.func_name == func_name:
                            test_suite.addTest(test_case.test_case)
                    else:
                        test_suite.addTest(test_case.test_case)

        elif func_name:
            test_suite = unittest.TestSuite()
            for test_cases in JacTestCheck.test_suite_path.values():
                for test_case in test_cases:
                    if test_case.func_name == func_name:
                        test_suite.addTest(test_case.test_case)

        runner = JacTextTestRunner(max_failures=maxfail, failfast=xit, verbosity=verb)
        result = runner.run(test_suite)
        if result.wasSuccessful():
            print("Passed successfully.")
        else:
            fails = len(result.failures)
            JacTestCheck.failcount += fails
            JacTestCheck.breaker = (
                (JacTestCheck.failcount >= maxfail) if maxfail else True
            )

    @staticmethod
    def add_test(filepath: str, func_name: str, test_func: Callable) -> None:
        """Create a new test."""
        if filepath and filepath.endswith(".test.jac"):
            filepath = filepath[:-9]
        elif filepath and filepath.endswith(".jac"):
            filepath = filepath[:-4]

        if filepath not in JacTestCheck.test_suite_path:
            JacTestCheck.test_suite_path[filepath] = []

        test_case = unittest.FunctionTestCase(test_func)
        JacTestCheck.test_suite_path[filepath].append(
            JacTestCheck.TestSuite(test_case=test_case, func_name=func_name)
        )
        JacTestCheck.test_suite.addTest(test_case)

    def __getattr__(self, name: str) -> object:
        """Make convenient check.Equal(...) etc."""
        return getattr(JacTestCheck.test_case, name)



"""Test registry pass."""

import os

from jaclang.compiler.compile import jac_file_to_pass
from jaclang.compiler.passes.main import RegistryPass
from jaclang.utils.test import TestCase


class RegistryPassTests(TestCase):
    """Test pass module."""

    # Need change
    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_registry_pass(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("registry.jac"), RegistryPass)
        self.assertFalse(state.errors_had)
        self.assertFalse(
            os.path.exists(
                os.path.join(
                    os.path.dirname(self.fixture_abs_path("registry.jac")),
                    "__jac_gen__",
                    "registry.registry.pkl",
                )
            )
        )
        self.assertIn("109", str(state.ir.to_dict()))



"""Living Workspace of Jac project."""

from __future__ import annotations

import asyncio
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Optional

import jaclang.compiler.absyntree as ast
from jaclang.compiler.compile import jac_str_to_pass
from jaclang.compiler.parser import JacParser
from jaclang.compiler.passes import Pass
from jaclang.compiler.passes.main.schedules import py_code_gen_typed
from jaclang.compiler.passes.tool import FuseCommentsPass, JacFormatPass
from jaclang.langserve.sem_manager import SemTokManager
from jaclang.langserve.utils import (
    add_unique_text_edit,
    collect_all_symbols_in_scope,
    collect_child_tabs,
    create_range,
    find_deepest_symbol_node_at_pos,
    find_index,
    gen_diagnostics,
    get_location_range,
    get_symbols_for_outline,
    parse_symbol_path,
)
from jaclang.vendor.pygls import uris
from jaclang.vendor.pygls.server import LanguageServer

import lsprotocol.types as lspt


class ModuleInfo:
    """Module IR and Stats."""

    def __init__(
        self,
        ir: ast.Module,
        impl_parent: Optional[ModuleInfo] = None,
    ) -> None:
        """Initialize module info."""
        self.ir = ir
        self.impl_parent: Optional[ModuleInfo] = impl_parent
        self.sem_manager = SemTokManager(ir=ir)
        self.is_modified: bool = False

    @property
    def uri(self) -> str:
        """Return uri."""
        return uris.from_fs_path(self.ir.loc.mod_path)


class JacLangServer(LanguageServer):
    """Class for managing workspace."""

    def __init__(self) -> None:
        """Initialize workspace."""
        super().__init__("jac-lsp", "v0.1")
        self.modules: dict[str, ModuleInfo] = {}
        self.executor = ThreadPoolExecutor()
        self.tasks: dict[str, asyncio.Task] = {}

    def update_modules(
        self, file_path: str, build: Pass, refresh: bool = False
    ) -> None:
        """Update modules."""
        if not isinstance(build.ir, ast.Module):
            self.log_error("Error with module build.")
            return
        keep_parent = (
            self.modules[file_path].impl_parent if file_path in self.modules else None
        )
        self.modules[file_path] = ModuleInfo(ir=build.ir, impl_parent=keep_parent)
        for p in build.ir.mod_deps.keys():
            uri = uris.from_fs_path(p)
            if file_path != uri:
                self.modules[uri] = ModuleInfo(
                    ir=build.ir.mod_deps[p],
                    impl_parent=self.modules[file_path],
                )

    def quick_check(self, file_path: str) -> bool:
        """Rebuild a file."""
        try:
            document = self.workspace.get_text_document(file_path)
            build = jac_str_to_pass(
                jac_str=document.source, file_path=document.path, schedule=[]
            )
            self.publish_diagnostics(
                file_path,
                gen_diagnostics(file_path, build.errors_had, build.warnings_had),
            )
            return len(build.errors_had) == 0
        except Exception as e:
            self.log_error(f"Error during syntax check: {e}")
            return False

    def deep_check(self, file_path: str, annex_view: Optional[str] = None) -> bool:
        """Rebuild a file and its dependencies."""
        try:
            start_time = time.time()
            document = self.workspace.get_text_document(file_path)
            if file_path in self.modules and (
                parent := self.modules[file_path].impl_parent
            ):
                return self.deep_check(
                    uris.from_fs_path(parent.ir.loc.mod_path), annex_view=file_path
                )
            build = jac_str_to_pass(
                jac_str=document.source,
                file_path=document.path,
                schedule=py_code_gen_typed,
            )
            self.update_modules(file_path, build)
            if discover := self.modules[file_path].ir.annexable_by:
                return self.deep_check(
                    uris.from_fs_path(discover), annex_view=file_path
                )

            self.publish_diagnostics(
                annex_view if annex_view else file_path,
                gen_diagnostics(
                    annex_view if annex_view else file_path,
                    build.errors_had,
                    build.warnings_had,
                ),
            )
            if annex_view:
                self.publish_diagnostics(
                    file_path,
                    gen_diagnostics(
                        file_path,
                        build.errors_had,
                        build.warnings_had,
                    ),
                )
            self.log_py(f"PROFILE: Deep check took {time.time() - start_time} seconds.")
            return len(build.errors_had) == 0
        except Exception as e:
            self.log_error(f"Error during deep check: {e}")
            return False

    async def launch_quick_check(self, uri: str) -> None:
        """Analyze and publish diagnostics."""
        await asyncio.get_event_loop().run_in_executor(
            self.executor, self.quick_check, uri
        )

    async def launch_deep_check(self, uri: str) -> None:
        """Analyze and publish diagnostics."""

        async def run_in_executor(
            func: Callable[[str, Optional[str]], bool],
            file_path: str,
            annex_view: Optional[str] = None,
        ) -> None:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(self.executor, func, file_path, annex_view)

        if uri in self.tasks and not self.tasks[uri].done():
            self.log_py(f"Canceling {uri} deep check...")
            self.tasks[uri].cancel()
            del self.tasks[uri]
        self.log_py(f"Analyzing {uri}...")
        task = asyncio.create_task(run_in_executor(self.deep_check, uri))
        self.tasks[uri] = task
        await task

    def get_completion(
        self, file_path: str, position: lspt.Position, completion_trigger: Optional[str]
    ) -> lspt.CompletionList:
        """Return completion for a file."""
        document = self.workspace.get_text_document(file_path)
        mod_ir = self.modules[file_path].ir
        current_line = document.lines[position.line]
        current_pos = position.character
        current_symbol_path = parse_symbol_path(current_line, current_pos)
        builtin_tab = mod_ir.sym_tab.kid[-1]
        completion_items = []

        node_selected = find_deepest_symbol_node_at_pos(
            mod_ir,
            position.line,
            position.character - 2,
        )
        mod_tab = mod_ir.sym_tab if not node_selected else node_selected.sym_tab
        current_symbol_table = mod_tab

        if completion_trigger == ".":
            if current_symbol_path:
                temp_tab = mod_tab
                for symbol in current_symbol_path:
                    if symbol == "self":
                        is_ability_def = (
                            temp_tab.owner
                            if isinstance(temp_tab.owner, ast.AbilityDef)
                            else temp_tab.owner.find_parent_of_type(ast.AbilityDef)
                        )
                        if not is_ability_def:
                            archi_owner = mod_tab.owner.find_parent_of_type(
                                ast.Architype
                            )
                            temp_tab = (
                                archi_owner._sym_tab
                                if archi_owner and archi_owner._sym_tab
                                else mod_tab
                            )
                            continue
                        else:
                            archi_owner = (
                                (
                                    is_ability_def.decl_link.find_parent_of_type(
                                        ast.Architype
                                    )
                                )
                                if is_ability_def.decl_link
                                else None
                            )
                            temp_tab = (
                                archi_owner.sym_tab
                                if archi_owner and archi_owner.sym_tab
                                else temp_tab
                            )
                            continue
                    symb = temp_tab.lookup(symbol)
                    if symb:
                        fetc_tab = symb.fetch_sym_tab
                        if fetc_tab:
                            temp_tab = fetc_tab
                        else:
                            temp_tab = (
                                symb.defn[0].type_sym_tab
                                if symb.defn[0].type_sym_tab
                                else temp_tab
                            )
                    else:
                        break
                completion_items += collect_all_symbols_in_scope(
                    temp_tab, up_tree=False
                )
                if (
                    isinstance(temp_tab.owner, ast.Architype)
                    and temp_tab.owner.base_classes
                ):
                    base = []
                    for base_name in temp_tab.owner.base_classes.items:
                        if isinstance(base_name, ast.Name) and base_name.sym:
                            base.append(base_name.sym)
                    for base_class_symbol in base:
                        if base_class_symbol.fetch_sym_tab:
                            completion_items += collect_all_symbols_in_scope(
                                base_class_symbol.fetch_sym_tab,
                                up_tree=False,
                            )

        else:
            if node_selected and (
                node_selected.find_parent_of_type(ast.Architype)
                or node_selected.find_parent_of_type(ast.AbilityDef)
            ):
                self_symbol = [
                    lspt.CompletionItem(
                        label="self", kind=lspt.CompletionItemKind.Variable
                    )
                ]
            else:
                self_symbol = []

            completion_items += (
                collect_all_symbols_in_scope(current_symbol_table)
                + self_symbol
                + collect_child_tabs(builtin_tab)
            )
        return lspt.CompletionList(is_incomplete=False, items=completion_items)

    def rename_module(self, old_path: str, new_path: str) -> None:
        """Rename module."""
        if old_path in self.modules and new_path != old_path:
            self.modules[new_path] = self.modules[old_path]
            del self.modules[old_path]

    def delete_module(self, uri: str) -> None:
        """Delete module."""
        if uri in self.modules:
            del self.modules[uri]

    def formatted_jac(self, file_path: str) -> list[lspt.TextEdit]:
        """Return formatted jac."""
        try:
            document = self.workspace.get_text_document(file_path)
            format = jac_str_to_pass(
                jac_str=document.source,
                file_path=document.path,
                target=JacFormatPass,
                schedule=[FuseCommentsPass, JacFormatPass],
            )
            formatted_text = (
                format.ir.gen.jac
                if JacParser not in [e.from_pass for e in format.errors_had]
                else document.source
            )
        except Exception as e:
            self.log_error(f"Error during formatting: {e}")
            formatted_text = document.source
        return [
            lspt.TextEdit(
                range=lspt.Range(
                    start=lspt.Position(line=0, character=0),
                    end=lspt.Position(
                        line=len(document.source.splitlines()) + 1, character=0
                    ),
                ),
                new_text=(formatted_text),
            )
        ]

    def get_hover_info(
        self, file_path: str, position: lspt.Position
    ) -> Optional[lspt.Hover]:
        """Return hover information for a file."""
        if file_path not in self.modules:
            return None
        token_index = find_index(
            self.modules[file_path].sem_manager.sem_tokens,
            position.line,
            position.character,
        )
        if token_index is None:
            return None
        node_selected = self.modules[file_path].sem_manager.static_sem_tokens[
            token_index
        ][3]
        value = self.get_node_info(node_selected) if node_selected else None
        if value:
            return lspt.Hover(
                contents=lspt.MarkupContent(
                    kind=lspt.MarkupKind.PlainText, value=f"{value}"
                ),
            )
        return None

    def get_node_info(self, node: ast.AstSymbolNode) -> Optional[str]:
        """Extract meaningful information from the AST node."""
        try:
            if isinstance(node, ast.NameAtom):
                node = node.name_of
            access = node.sym.access.value + " " if node.sym else None
            node_info = (
                f"({access if access else ''}{node.sym_category.value}) {node.sym_name}"
            )
            if node.name_spec.clean_type:
                node_info += f": {node.name_spec.clean_type}"
            if isinstance(node, ast.AstSemStrNode) and node.semstr:
                node_info += f"\n{node.semstr.value}"
            if isinstance(node, ast.AstDocNode) and node.doc:
                node_info += f"\n{node.doc.value}"
            if isinstance(node, ast.Ability) and node.signature:
                node_info += f"\n{node.signature.unparse()}"
            self.log_py(f"mypy_node: {node.gen.mypy_ast}")
        except AttributeError as e:
            self.log_warning(f"Attribute error when accessing node attributes: {e}")
        return node_info.strip()

    def get_outline(self, file_path: str) -> list[lspt.DocumentSymbol]:
        """Return document symbols for a file."""
        if file_path in self.modules and (
            root_node := self.modules[file_path].ir._sym_tab
        ):
            return get_symbols_for_outline(root_node)
        return []

    def get_definition(
        self, file_path: str, position: lspt.Position
    ) -> Optional[lspt.Location]:
        """Return definition location for a file."""
        if file_path not in self.modules:
            return None
        token_index = find_index(
            self.modules[file_path].sem_manager.sem_tokens,
            position.line,
            position.character,
        )
        if token_index is None:
            return None
        node_selected = self.modules[file_path].sem_manager.static_sem_tokens[
            token_index
        ][3]
        if node_selected:
            if (
                isinstance(node_selected, ast.Name)
                and node_selected.parent
                and isinstance(node_selected.parent, ast.ModulePath)
            ):
                spec = node_selected.parent.abs_path
                if spec:
                    spec = spec[5:] if spec.startswith("File:") else spec
                    return lspt.Location(
                        uri=uris.from_fs_path(spec),
                        range=lspt.Range(
                            start=lspt.Position(line=0, character=0),
                            end=lspt.Position(line=0, character=0),
                        ),
                    )
                else:
                    return None
            elif node_selected.parent and isinstance(
                node_selected.parent, ast.ModuleItem
            ):
                path = (
                    node_selected.parent.abs_path
                    or node_selected.parent.from_mod_path.abs_path
                )
                try:  # TODO: Get rid of this when 'from' import is fixed
                    loc_range = tuple(
                        loc - 1 if loc > 0 else loc
                        for loc in get_location_range(node_selected.parent)
                    )
                except ValueError:
                    loc_range = (0, 0, 0, 0)

                if path and loc_range:
                    path = path[5:] if path.startswith("File:") else path
                    return lspt.Location(
                        uri=uris.from_fs_path(path),
                        range=lspt.Range(
                            start=lspt.Position(
                                line=loc_range[0], character=loc_range[1]
                            ),
                            end=lspt.Position(
                                line=loc_range[2], character=loc_range[3]
                            ),
                        ),
                    )
            elif isinstance(node_selected, ast.ElementStmt):
                return None
            decl_node = (
                node_selected.parent.body.target
                if node_selected.parent
                and isinstance(node_selected.parent, ast.AstImplNeedingNode)
                and isinstance(node_selected.parent.body, ast.AstImplOnlyNode)
                else (
                    node_selected.sym.decl
                    if (node_selected.sym and node_selected.sym.decl)
                    else node_selected
                )
            )
            decl_uri = uris.from_fs_path(decl_node.loc.mod_path)
            try:
                decl_range = create_range(decl_node.loc)
            except ValueError:
                return None
            decl_location = lspt.Location(
                uri=decl_uri,
                range=decl_range,
            )

            return decl_location
        else:
            return None

    def get_references(
        self, file_path: str, position: lspt.Position
    ) -> list[lspt.Location]:
        """Return references for a file."""
        if file_path not in self.modules:
            return []
        index1 = find_index(
            self.modules[file_path].sem_manager.sem_tokens,
            position.line,
            position.character,
        )
        if index1 is None:
            return []
        node_selected = self.modules[file_path].sem_manager.static_sem_tokens[index1][3]
        if node_selected and node_selected.sym:
            list_of_references: list[lspt.Location] = [
                lspt.Location(
                    uri=uris.from_fs_path(node.loc.mod_path),
                    range=create_range(node.loc),
                )
                for node in node_selected.sym.uses
            ]
            return list_of_references
        return []

    def rename_symbol(
        self, file_path: str, position: lspt.Position, new_name: str
    ) -> Optional[lspt.WorkspaceEdit]:
        """Rename a symbol in a file."""
        if file_path not in self.modules:
            return None
        index1 = find_index(
            self.modules[file_path].sem_manager.sem_tokens,
            position.line,
            position.character,
        )
        if index1 is None:
            return None
        node_selected = self.modules[file_path].sem_manager.static_sem_tokens[index1][3]
        if node_selected and node_selected.sym:
            changes: dict[str, list[lspt.TextEdit]] = {}
            for node in [
                *node_selected.sym.uses,
                node_selected.sym.defn[0],
            ]:
                key = uris.from_fs_path(node.loc.mod_path)
                new_edit = lspt.TextEdit(
                    range=create_range(node.loc),
                    new_text=new_name,
                )
                add_unique_text_edit(changes, key, new_edit)
            return lspt.WorkspaceEdit(changes=changes)
        return None

    def get_semantic_tokens(self, file_path: str) -> lspt.SemanticTokens:
        """Return semantic tokens for a file."""
        if file_path not in self.modules:
            return lspt.SemanticTokens(data=[])
        return lspt.SemanticTokens(data=self.modules[file_path].sem_manager.sem_tokens)

    def log_error(self, message: str) -> None:
        """Log an error message."""
        self.show_message_log(message, lspt.MessageType.Error)
        self.show_message(message, lspt.MessageType.Error)

    def log_warning(self, message: str) -> None:
        """Log a warning message."""
        self.show_message_log(message, lspt.MessageType.Warning)
        self.show_message(message, lspt.MessageType.Warning)

    def log_info(self, message: str) -> None:
        """Log an info message."""
        self.show_message_log(message, lspt.MessageType.Info)
        self.show_message(message, lspt.MessageType.Info)

    def log_py(self, message: str) -> None:
        """Log a message."""
        logging.info(message)



"""Test ast build pass module."""

import ast as ast3
from difflib import unified_diff


from jaclang.compiler.compile import jac_file_to_pass, jac_str_to_pass
from jaclang.compiler.passes.main import PyastGenPass
from jaclang.compiler.passes.main.schedules import py_code_gen as without_format
from jaclang.compiler.passes.tool import JacFormatPass
from jaclang.utils.test import AstSyncTestMixin, TestCaseMicroSuite


class JacUnparseTests(TestCaseMicroSuite, AstSyncTestMixin):
    """Test pass module."""

    TargetPass = JacFormatPass

    def test_double_unparse(self) -> None:
        """Parse micro jac file."""
        try:
            code_gen_pure = jac_file_to_pass(
                self.examples_abs_path("manual_code/circle.jac"),
                target=PyastGenPass,
                schedule=without_format,
            )
            x = code_gen_pure.ir.unparse()
            y = code_gen_pure.ir.unparse()
            self.assertEqual(x, y)
        except Exception as e:
            print("\n".join(unified_diff(x.splitlines(), y.splitlines())))
            raise e

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        code_gen_pure = jac_file_to_pass(
            self.fixture_abs_path(filename),
            target=PyastGenPass,
            schedule=without_format,
        )
        before = ast3.dump(code_gen_pure.ir.gen.py_ast[0], indent=2)
        x = code_gen_pure.ir.unparse()
        # print(x)
        # print(f"Testing {code_gen_pure.ir.name}")
        # print(code_gen_pure.ir.pp())
        code_gen_jac = jac_str_to_pass(
            jac_str=x,
            file_path=filename,
            target=PyastGenPass,
            schedule=without_format,
        )
        after = ast3.dump(code_gen_jac.ir.gen.py_ast[0], indent=2)
        if "circle_clean_tests.jac" in filename:
            self.assertEqual(
                len(
                    [
                        i
                        for i in unified_diff(
                            before.splitlines(), after.splitlines(), n=0
                        )
                        if "test" not in i
                    ]
                ),
                5,
            )
        else:
            self.assertEqual(
                len("\n".join(unified_diff(before.splitlines(), after.splitlines()))),
                0,
            )


JacUnparseTests.self_attach_micro_tests()
